import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import netscape.server.applet.*;
import oracle.jdbc.driver.*;
import java.sql.*;

import com.ms.xml.om.*;
import com.ms.xml.util.*;
import com.ms.xml.parser.*;


public class TDLServlet extends HttpServlet {
	
	static AcceptedTagList tagCollection = new AcceptedTagList();
    static Hashtable templateCollection = new Hashtable();
    static Cache outputCache = new Cache();
    
    
	public void init (ServletConfig config) throws ServletException {
    
    	super.init(config);
        ExceptionHandler.initLogFile();
		String driverName = Constants.DatabaseDriver;
        try {
        	Driver driver = (Driver)Class.forName(driverName).newInstance();
        } catch (Exception e) {
		    ExceptionHandler.handleException(e,"DBDriver exception");
		}
		String tagFileName = Constants.TagConfigurationFileName;
		String templateFileName = Constants.TemplateConfigurationFileName;
        readTagCollection (tagFileName, tagCollection);
        readTemplateCollection (templateFileName, templateCollection);
        ExceptionHandler.debugInfo("Initialization complete. Waiting for requests.");
	}

	public void doGet (HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
        Document xmlDocument = new Document();
        ServletOutputStream out = null;
        String ip = req.getRemoteAddr();
        try {
            out = res.getOutputStream();
        } catch (Exception e){}
  	    res.setContentType("text/html");

        String pathInfo = req.getPathInfo();
        String getQueryString = req.getQueryString();

        if (pathInfo.indexOf("/presentation") != -1) {
            redirectPresentation(pathInfo, out, ip);
            return;
        }

   	    if (pathInfo.equals("/config")) {
       	    try {
       	       String fNumber = getQueryString;
       	       int extra = 0;
       	       if ( (getQueryString.substring(0,2).equals("33")) || (getQueryString.substring(0,2).equals("24"))) {
       	            extra = Integer.parseInt(getQueryString.substring(2,getQueryString.length()));
       	            getQueryString = getQueryString.substring(0,2);
       	       }
               returnConfigPage (Integer.parseInt(getQueryString),req.getServerName(), out, extra);
               return;
            } catch (NumberFormatException e) {
               showInfoScreen(out,0);
               return;
            }
        }
        String result;
        String cacheKey = pathInfo + getQueryString;
        if ( !(outputCache.containsKey(cacheKey)) || !(outputCache.isCacheActive()) ) {

            if (pathInfo==null) { showInfoScreen(out,0); return; }
            pathInfo = pathInfo.substring(1,pathInfo.length());
            String templateId = URIUtil.urlUnescape(pathInfo);
            String templateFileName = "";
            int slashLoc;
            if ((slashLoc=pathInfo.indexOf("/"))!=-1) {
              	templateId = URIUtil.urlUnescape(pathInfo.substring(0,slashLoc));
                templateFileName = URIUtil.urlUnescape(pathInfo.substring(slashLoc+1,pathInfo.length()));
            }
            if (templateFileName.trim().equals("")) templateFileName = Constants.DefaultTemplateLoginPageFileName;
            if (!templateCollection.containsKey(templateId)) { showInfoScreen(out,1); return; }
            
            Template template = ((Template)templateCollection.get(templateId));
            try {
                template.loadProperties();
            } catch (Exception e) {
                ExceptionHandler.handleException(e,"loading properties");
            }
            if (getQueryString!=null) template.setQueryParameters(getQueryString);
            
//            showProperties(template,out);
            Hashtable localProperties = template.cloneTemplateProperties();
            InputStream is = template.getTemplateFileStream(templateFileName);
            Vector tags = new Vector();
            Enumeration en = tagCollection.keys();
            while (en.hasMoreElements()) {
                tags.addElement(en.nextElement());
            }
            tags.addElement("?XML");

            try {
                WWInlineInputStream iis = new WWInlineInputStream(is, template);
                XMLEncodeInputStream xmlEIS = new XMLEncodeInputStream(iis,tags);
                xmlDocument.load(xmlEIS);
            }
            catch (Exception e) {
                ExceptionHandler.handleException(e, "XML not found.");
            }

            Connection connectionForGet=null;
            try {
                connectionForGet = DriverManager.getConnection((String)template.getProperty(Constants.ConnectionStringProperty) );
            } catch (SQLException e ){
    		    ExceptionHandler.handleException(e,"redirectPresentation");   
    		    return;
            }

  
            result = parse(localProperties, tagCollection, template, localProperties, ip, xmlDocument, connectionForGet);
  
            outputCache.put(cacheKey,result);
  
            try {
                if (connectionForGet != null) connectionForGet.close();
            } catch (SQLException e) {
                ExceptionHandler.handleException(e, "closeCon");
            }
              
        }
        else {
            result = (String)outputCache.get(cacheKey);
        }
        decodeCacheAndFlush (result,out);
        
	}

	public void doPost (HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
   	    Template t;

	    ServletOutputStream out = null;
        try { 
            out = res.getOutputStream(); 
        } catch (Exception e){
            ExceptionHandler.handleException(e,"doPost.getOutputStream");    
        }
  	    res.setContentType("text/html");
        
   	    String whatToDo = req.getParameter("todo");
        ExceptionHandler.debugInfo("What to do:"+whatToDo);
   	    if (whatToDo.equals("login")) {
            String ip = req.getRemoteAddr();
            String templateId = req.getParameter ("template");
            String uname = req.getParameter ("user");
            String upass = req.getParameter ("pass");

            if (!templateCollection.containsKey(templateId)) { showInfoScreen(out,1); return; }

            Template template = ((Template)templateCollection.get(templateId));
            redirectLogin (checkAuth (uname, upass, template), out, templateId, ip);
   	    }
   	    
   	    //general cache
   	    if (whatToDo.equals("expireOutCache")) {
   	        outputCache.expireCache();
   	        cacheInfo(out);
   	    }
   	    if (whatToDo.equals("expireTemplateCache")) {
            Enumeration en = templateCollection.keys();
            while (en.hasMoreElements()) {
                t = (Template)templateCollection.get(en.nextElement());
       	        t.expireTemplateCache();
   	        }
   	        cacheInfo(out);
   	    }
   	    
   	    //output cache
   	    if (whatToDo.equals("ocacheoff")) {
   	        if (outputCache.isCacheActive()) outputCache.setCacheInactive();
   	        else outputCache.setCacheActive();
   	        outputCacheControl(out);
   	    }
   	    if (whatToDo.equals("ocacheexp")) {
   	        outputCache.expireCache();
   	        outputCacheControl(out);
   	    }
   	    if (whatToDo.equals("ocacheapp")) {
            String time = req.getParameter ("maxTime");
            String hits = req.getParameter ("maxHits");

   	        outputCache.setMaxTime(Integer.parseInt(time));
   	        outputCache.setMaxHits(Integer.parseInt(hits));
   	        outputCacheControl(out);
   	    }
   	    
   	    //template cache
   	    if (whatToDo.equals("tcacheoff")) {
            String tname = req.getParameter ("templ");
            int extra = Integer.parseInt(req.getParameter ("extra"));
   	        t = (Template)templateCollection.get(tname);
   	        
   	        if (t.isTemplateCacheActive()) t.setTCacheInactive();
   	        else t.setTCacheActive();
   	        templateCacheControl(out,extra);
   	    }
   	    if (whatToDo.equals("tcacheexp")) {
            String tname = req.getParameter ("templ");
            int extra = Integer.parseInt(req.getParameter ("extra"));
   	        t = (Template)templateCollection.get(tname);

   	        t.expireTemplateCache();
   	        templateCacheControl(out,extra);
   	    }
   	    if (whatToDo.equals("tcacheapp")) {
            String time = req.getParameter ("maxTime");
            String hits = req.getParameter ("maxHits");
            String tname = req.getParameter ("templ");
            int extra = Integer.parseInt(req.getParameter ("extra"));
   	        t = (Template)templateCollection.get(tname);

   	        t.setTMaxTime(Integer.parseInt(time));
   	        t.setTMaxHits(Integer.parseInt(hits));
   	        templateCacheControl(out,extra);
   	    }
    }



    

//***********************************************************************
//end of main methods
//***********************************************************************

    private int getPid (String pname, Connection cn) {
        int pid = -1;
        try {
            String query = "SELECT PID FROM PRESENTATION WHERE NAME = '"+pname+"'";
       		Statement sqlStatement = cn.createStatement();
          	ResultSet rs = sqlStatement.executeQuery(query);
          	if (rs.next()) {
                pid = rs.getInt ("PID");
          	}
            rs.close(); 
            sqlStatement.close();
        } catch (SQLException e ){
		    ExceptionHandler.handleException(e,"getPid");   
        }
        return pid;
    }


    private int getUsid (int pid, String uname, String ip, Connection cn) {
        String luname = "";
        int usid = -1;
        if (uname.equals("")) {
            try {
                String query = "SELECT VALUE FROM PRES_GENERAL_PROPERTY WHERE PRES_PID = "+pid+" AND GEN_PROP_PRID = (SELECT PRID FROM GENERAL_PROPERTY WHERE SEARCH_NAME='"+Constants.GeneralPropUserName+"')";

           		Statement sqlStatement = cn.createStatement();
              	ResultSet rs = sqlStatement.executeQuery(query);
              	if (rs.next()) {
                    luname = rs.getString ("VALUE");
              	}
                rs.close(); 
                sqlStatement.close();
                if (luname.equals("")) {
                    query = "SELECT DEF_VALUE FROM GENERAL_PROPERTY WHERE SEARCH_NAME='"+Constants.GeneralPropUserName+"'";

               		sqlStatement = cn.createStatement();
                  	rs = sqlStatement.executeQuery(query);
                  	if (rs.next()) {
                        luname = rs.getString ("DEF_VALUE");
                  	}
                    rs.close(); 
                    sqlStatement.close();
                }
                if (luname.equals(Constants.NoAccessGeneralProperty)) {
                    return usid;
                }

                query = "SELECT USID FROM USERS WHERE LOGIN_NAME='"+luname+"'";
           		sqlStatement = cn.createStatement();
              	rs = sqlStatement.executeQuery(query);
              	if (rs.next()) {
                    usid = rs.getInt ("USID");
              	}
                rs.close(); 
                sqlStatement.close();
                usid = scramble (usid, ip);
            } catch (SQLException e ){
    		    ExceptionHandler.handleException(e,"getUsid");   
            }
            return usid;    
        }
        else {
            boolean isAny = false;
            try {
                String query = "SELECT USID FROM USERS WHERE LOGIN_NAME='"+uname+"'";
           		Statement sqlStatement = cn.createStatement();
              	ResultSet rs = sqlStatement.executeQuery(query);
              	if (rs.next()) {
                    usid = rs.getInt ("USID");
              	}
                rs.close(); 
                sqlStatement.close();
                query = "SELECT ACCESS_MODE FROM ACCESS_TO_FOILWORLD WHERE USER_UID = "+usid+" AND FOILWORLD_WID IN (SELECT CONTAINER_WID FROM FW_CONTAIN_PR WHERE COMPONENT_PID = "+pid+")";
           		sqlStatement = cn.createStatement();
              	rs = sqlStatement.executeQuery(query);
              	while (rs.next()) {
                    String tmp = rs.getString("ACCESS_MODE");
                    if (tmp.indexOf(Constants.AccessModeReadRight) != -1) isAny = true; 
              	}
                rs.close(); 
                sqlStatement.close();
                if (!isAny) {
                    query = "SELECT WID FROM FOILWORLD WHERE OWNER_UID = "+usid+" AND BUFFER_OWNER_UID IS NULL AND WID IN (SELECT CONTAINER_WID FROM FW_CONTAIN_PR WHERE COMPONENT_PID = "+pid+")";
               		sqlStatement = cn.createStatement();
                  	rs = sqlStatement.executeQuery(query);
                  	if (rs.next()) {
                        isAny = true; 
                  	}
                    rs.close(); 
                }
            } catch (SQLException e ){
    		    ExceptionHandler.handleException(e,"getUsid");   
            }
    
            if (isAny) return scramble(usid,ip);
            else return -1;
        }
    }


    private void redirectPresentation (String pathInfo, ServletOutputStream out, String ip) {
        pathInfo = pathInfo.substring (1,pathInfo.length());
        StringTokenizer st = new StringTokenizer (pathInfo, "/");
        String tmp = st.nextToken();
        String pname = st.nextToken();
        String tname = st.nextToken();
        String uname = "";
        if (st.hasMoreTokens()) uname = st.nextToken();
        
        Template template = ((Template)templateCollection.get(tname));
        if (template == null) {
//no such template
            showInfoScreen(out,1);
            return;
        }
        try {
            template.loadProperties();
        } catch (Exception e) {
            ExceptionHandler.handleException(e,"loading properties");
        }
        Connection cn=null;
        try {
            cn = DriverManager.getConnection((String)template.getProperty(Constants.ConnectionStringProperty) );
        } catch (SQLException e ){
		    ExceptionHandler.handleException(e,"redirectPresentation");   
        }
        
        int pid = getPid (pname, cn);
        if (pid == -1) {
//no such presentation
//            ExceptionHandler.debugInfo("pid");
            showInfoScreen(out,2);
            return;
        }
        int usid = getUsid (pid, uname, ip, cn);
        if (usid == -1) {
//user has no rights
//            ExceptionHandler.debugInfo("usid");
            showInfoScreen(out,3);
            return;
        }
        String url = "/"+(String)template.getProperty(Constants.ServletDirProperty)+"/"+(String)template.getProperty(Constants.ServletNameProperty)+"/"+tname+"/tempfullindex.tdl?PID="+pid+"&USID="+usid;
        String res = "<HTML><HEAD><META HTTP-EQUIV=\"Refresh\" CONTENT=\"0; URL=";
   		res += url;
   		res += "\"><TITLE>Moving to presentation</TITLE></HEAD><BODY BGCOLOR=\"#FEF5C1\" LINK=\"Black\" VLINK=\"Black\" ALINK=\"Black\"><TT>Please wait while the presentation <A HREF=\""+url+"\">"+pname+"</A> is loading...</TT></BODY></HTML>";

        printOut (res,out);
    }


    private int scramble (int usid, String ip) {
        int res = 1;
        StringTokenizer st = new StringTokenizer(ip,".");
        String tok = st.nextToken();
        while (st.hasMoreTokens()) {
            tok = st.nextToken();
            res *= Integer.parseInt(tok);
        }
//        return res + usid;
        return usid;
    }
    
    private int checkAuth (String uname, String upass, Template templ) {
        int res  = -1;
        try {
            Connection cn = DriverManager.getConnection((String)templ.getProperty(Constants.ConnectionStringProperty) );
            String query = "SELECT USID, USER_PRIVILEGES FROM USERS WHERE LOGIN_NAME = '"+uname+"' AND PASSWORD = '"+upass+"'";
       		Statement sqlStatement = cn.createStatement();
          	ResultSet rs = sqlStatement.executeQuery(query);
          	if (rs.next()) {
          	    String priv = rs.getString ("USER_PRIVILEGES");
          	    if (priv.indexOf ("_CAN_USE_WW") != -1) res = rs.getInt("USID");
          	}
            rs.close(); 
            sqlStatement.close();
        } catch (SQLException e) {
		    ExceptionHandler.handleException(e,"checkAuth");   
        }
        return res;
    }
    
    private void redirectLogin (int usid, ServletOutputStream out, String templateId, String ip) {
        usid = scramble (usid,ip);
        if (usid != -1) {
            String s = Constants.DefaultTemplateStartingPageFileName;
            printHead2("",out, templateId, usid);
            printOut("<CENTER><H2>User logged in</H2></CENTER><HR>",out);
            printOut("Use the following link to enter template: ",out);
            printOut("<A HREF=\""+s+"?USID="+usid+"\"> Enter the template </A>",out);
            printTail(out);    
        }
        else {
            String sl = Constants.DefaultTemplateLoginPageFileName;
            printHead("",out);
            printOut("<CENTER><H2>Login incorrect</H2></CENTER><HR>",out);
            printOut("Use the following link to retry: ",out);
            printOut("<A HREF=\""+templateId+"/"+sl+"\"> Back to Login Page </A>",out);
            printTail(out);    
        }
    }


	
	static private void readTagCollection (String fileName, Hashtable hash) {
		String in,k,v;
        try {
		    BufferedReader inTemplateFile = new BufferedReader (new FileReader (fileName));
    		while ((in = inTemplateFile.readLine ())!= null) {
    		    k = in.substring(0,in.indexOf("|")).trim();
    		    v = in.substring( in.indexOf("|")+1,in.length()).trim();
    		    if (!(k.equals("") || v.equals(""))) 
    		        hash.put(k,v);
	    	}
		} catch (Exception e) {
		    ExceptionHandler.handleException(e,"readTagCollection");   
		}
	}

	static private void readTemplateCollection (String fileName, Hashtable hash) {
		String in;
        try {
		    BufferedReader inTemplateFile = new BufferedReader (new FileReader (fileName));
    		while ((in = inTemplateFile.readLine ())!= null) {
                Template t = new Template(tagCollection);     		
                t.addToProperties(Constants.TemplateIDProperty, in.substring( 0,in.indexOf("|")));
                t.addToProperties(Constants.TemplateLocationProperty, in.substring( in.indexOf("|")+1,in.length()));
                t.addToProperties(Constants.IFLevelProperty,"0"); 
                t.addToProperties(Constants.ServletNameProperty,"TDLServlet");
                t.addToProperties(Constants.ServletDirProperty,"servlet");
	    		hash.put(in.substring(0,in.indexOf("|")),t);
	    	}
		} catch (Exception e) {
		    ExceptionHandler.handleException(e,"readTemplateCollection");   
		}
	}

    private String parse(Hashtable templateProperties, AcceptedTagList tagList, Template templ, Hashtable localProperties, String ip, Document xmlDocument, Connection connectionForGet) {
        TDLRoot tdlRoot = new TDLRoot(); 
        tdlRoot.setTagCollection(tagList);
        tdlRoot.setClientIP(ip);
        tdlRoot.setConn(connectionForGet);
        return(tdlRoot.parseTagTree(templ, xmlDocument, localProperties));
    }
	
	
	private void decodeCacheAndFlush (String res, ServletOutputStream out) {
	    String line;
	    BufferedReader in = new BufferedReader (new XMLDecodeReader(new StringReader(res)));
        try {
    	    while ((line = in.readLine())!=null) out.println(line);   
	    } catch (Exception e) {
	        ExceptionHandler.handleException (e,"flush&Decode");   
	    }
	}
	

	private void printToOutStream (String s, ServletOutputStream out) {
       try {
      	   out.println(s+"<BR>");
      	}
      	catch (IOException e)
      	{
      	    ExceptionHandler.handleException(e,"printToOutputStream");
      	}
	}

    private void printOut (String s, ServletOutputStream out) { 
       try {
      	   out.println(s);
      	}
      	catch (IOException e)
      	{
      	    ExceptionHandler.handleException(e,"printToOutputStream");
      	}
	}


    private void printHead (String t, ServletOutputStream out) {
        printOut("<HTML><HEAD><TITLE>"+t+"</TITLE></HEAD><BODY BGCOLOR=\"#FEF5C1\" LINK=\"Black\" VLINK=\"Black\" ALINK=\"Black\">",out);        
    }
    
    private void printHead2 (String t, ServletOutputStream out, String templateId,int usid) {
        String s = Constants.DefaultTemplateStartingPageFileName;
        printOut("<HTML><HEAD><META HTTP-EQUIV=\"Refresh\" CONTENT=\"0; URL="+templateId+"/"+s+"?USID="+usid+"\"><TITLE>"+t+"</TITLE></HEAD><BODY BGCOLOR=\"#FEF5C1\" LINK=\"Black\" VLINK=\"Black\" ALINK=\"Black\">",out);        
    }
    
    private void printTail (ServletOutputStream out) {
        printOut("</BODY></HTML>",out);
    }
    
	private void showInfoScreen(ServletOutputStream out, int i) {
        printHead ("Servlet error",out);
        switch (i) {
            case 0:printOut("<B><TT>Servlet error: </B>Wrong parameters</TT><P>",out);break;
            case 1:printOut("<B><TT>Servlet error: </B>No such template</TT><P>",out);break;
            case 2:printOut("<B><TT>Servlet error: </B>No such presentation</TT><P>",out);break;
            case 3:printOut("<B><TT>Servlet error: </B>No rights for presentation</TT><P>",out);break;
        }
      	printTail(out);
	}

    private void servletInfo (String sN, ServletOutputStream out) {
        Runtime rt = Runtime.getRuntime();
        
        printHead("",out);
        printOut("<CENTER><H2>Servlet information</H2></CENTER><HR>",out);
        printOut("<TT><B>General info</B></TT><P>",out);
        printOut("<TT>Servlet name: <I>TDL servlet</I></TT><BR>",out);
        printOut("<TT>Servlet version: <I>1.0</I></TT><BR>",out);
        printOut("<TT>Servlet authors: <I>KW&VV, 1998</I></TT><BR>",out);
        printOut("<TT>Servlet description: <I>TDL servlet parser for dynamic templates</I></TT><P>",out);
        printOut("<TT><B>System resources</B></TT><P>",out);
        printOut("<TT>Hostname: <I>"+sN+"</I></TT><BR>",out);
        printOut("<TT>System memory free: <I>"+rt.freeMemory()+"</I></TT><BR>",out);
        printOut("<TT>Java VM memory free: <I>"+rt.totalMemory()+"</I></TT><BR>",out);
        printTail(out);
    } 

    private void listKnownTags (ServletOutputStream out) {
        printHead("",out);
        printOut("<CENTER><H2>Known Tags</H2></CENTER><HR>",out);
        printOut("<TABLE BORDER=1>",out);
        printOut("<TR><TD BGCOLOR=\"#FF0000\"><TT><B>Tag name:</B></TT></TD>",out);
        printOut("<TD><TT><B>Tag class:</B></TT></TD></TR>",out);
        printOut("<TR><TD BGCOLOR=\"#FF8000\"><TT>INLINE</TT></TD>",out);
        printOut("<TD><TT>Internal</TT></TD></TR>",out);
        MethodInvocation mI;
        Enumeration en = tagCollection.keys();
        while (en.hasMoreElements()) {
                String el = (String)en.nextElement();
                mI = (MethodInvocation)tagCollection.get(el);
                printOut("<TR><TD BGCOLOR=\"#FF8000\"><TT>"+el+"</TT></TD>",out);
                printOut("<TD><TT>"+mI.className+"."+mI.methodName+"(",out);
                for (int i=0;i<mI.numberOfArguments;i++) {
                    printOut(mI.argumentList[i],out);
                    if (i!=mI.numberOfArguments-1) printOut(",",out);
                }
                printOut(")</TT></TD></TR>",out);
        }
        printOut("</TABLE>",out);
        printTail(out);    
    }

    private void cacheInfo (ServletOutputStream out) {
        printHead("",out);
        printOut("<CENTER><H2>Cache information</H2></CENTER><HR>",out);
        printOut("<TT><B>Output cache</B></TT><P>",out);
        String answer = outputCache.isCacheActive()?"YES":"NO";
        printOut("<TT>Is active: <I>"+answer+"</I></TT><BR>",out);
        Enumeration en = outputCache.keys();
        String tmpS;
        int cs = 0;
        while (en.hasMoreElements()) {
            tmpS = (String)outputCache.get(en.nextElement());
            if (tmpS!=null) cs+=tmpS.length();
        }
        int nok = outputCache.size();
        printOut("<TT>Cache size: <I>"+cs+"</I></TT><BR>",out);
        printOut("<TT>Number of entries: <I>"+nok+"</I></TT><BR>",out);
        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"Expire now\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"expireOutCache\"></FORM><P>",out);

        printOut("<TT><B>Template file cache</B></TT><P>",out);
        boolean ca = false;
        Template t;
        nok = 0;
        cs = 0;
        en = templateCollection.keys();
        while (en.hasMoreElements()) {
             t = (Template)templateCollection.get(en.nextElement());
             cs += t.getTemplateCacheSize();
             nok += t.getNoOfKeys();
             ca = ca || t.isTemplateCacheActive();
        }
        answer = ca?"YES":"NO";
        printOut("<TT>Is active: <I>"+answer+"</I></TT><BR>",out);
        printOut("<TT>Cache size: <I>"+cs+"</I></TT><BR>",out);
        printOut("<TT>Number of entries: <I>"+nok+"</I></TT><BR>",out);
        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"Expire now\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"expireTemplateCache\"></FORM><P>",out);
        
        printTail(out);
    }
    
    private void outputCacheControl (ServletOutputStream out) {
        printHead("",out);
        printOut("<CENTER><H2>Output cache control</H2></CENTER><HR>",out);
        String answer = outputCache.isCacheActive()?"YES":"NO";
        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<TT>Cache active: <I><B>"+answer+"</B></I></TT> &nbsp;&nbsp;",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"On/Off\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"ocacheoff\"></FORM><P><HR>",out);

        Enumeration en = outputCache.keys();
        String tmpS;
        int cs = 0;
        while (en.hasMoreElements()) {
            tmpS = (String)outputCache.get(en.nextElement());
            if (tmpS!=null) cs+=tmpS.length();
        }
        int nok = outputCache.size();
        printOut("<TABLE><TR><TD><TT>Cache size: <I>"+cs+"</I></TT><P>",out);
        printOut("<TT>Number of entries: <I>"+nok+"</I></TT><BR></TD><TD>",out);
        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"Expire now\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"ocacheexp\"></FORM></TD></TR></TABLE>",out);

        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<TT>Cache timeout: <I><INPUT TYPE=\"TEXT\" NAME=\"maxTime\" VALUE=\""+outputCache.getMaxTime()+"\" SIZE=8></I> (sec.)</TT><BR>",out);
        printOut("<TT>Cache maximum hits: <I><INPUT TYPE=\"TEXT\" NAME=\"maxHits\" VALUE=\""+outputCache.getMaxHits()+"\" SIZE=8></I></TT><BR>",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"Apply\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"ocacheapp\"></FORM><P>",out);

        
        printTail(out);
    }

    private void printScript (String topage, ServletOutputStream out) {
        printOut("<SCRIPT LANGUAGE=\"JavaScript1.2\">",out);
        printOut("function gotoPage () {",out);
		printOut("top.fright.location=\"http://taurus7/servlet/TDLServlet/config?"+topage+"\"+document.forselect.tempname.selectedIndex",out);
	    printOut("}",out);
        printOut("</SCRIPT>",out);
    }
    

    private void templateCacheControl (ServletOutputStream out, int extra) {
        printHead("",out);
        printScript("33",out);
        printOut("<CENTER><H2>Template cache control</H2></CENTER><HR>",out);
        printOut("<FORM NAME=\"forselect\"><SELECT NAME=\"tempname\" onChange=\"gotoPage()\">",out);
        Enumeration en = templateCollection.keys();
        int i=0;        
        String thisName = "";
        String currentName = "";
        while (en.hasMoreElements()) {
            thisName = (String)en.nextElement();
            if (i == extra) {
                printOut("<OPTION VALUE=\""+i+"\" SELECTED> "+thisName,out);
                currentName = thisName;
            }
            else {
                printOut("<OPTION VALUE=\""+i+"\"> "+thisName,out);
            }
            i++;
        }
        printOut("</SELECT></FORM><HR>",out);
        printOut("<TT><B>Cache control for template : "+currentName+"</B></TT><P>",out);
        
        
        Template template = (Template)templateCollection.get(currentName);
        
        
        String answer = template.isTemplateCacheActive()?"YES":"NO";
        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<TT>Cache active: <I><B>"+answer+"</B></I></TT> &nbsp;&nbsp;",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"On/Off\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"tcacheoff\"><INPUT TYPE=HIDDEN NAME=\"templ\" VALUE=\""+currentName+"\"><INPUT TYPE=HIDDEN NAME=\"extra\" VALUE=\""+extra+"\"></FORM><P><HR>",out);

        printOut("<TABLE><TR><TD><TT>Cache size: <I>"+template.getTemplateCacheSize()+"</I></TT><P>",out);
        printOut("<TT>Number of entries: <I>"+template.getNoOfKeys()+"</I></TT><BR></TD><TD>",out);
        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"Expire now\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"tcacheexp\"><INPUT TYPE=HIDDEN NAME=\"templ\" VALUE=\""+currentName+"\"><INPUT TYPE=HIDDEN NAME=\"extra\" VALUE=\""+extra+"\"></FORM></TD></TR></TABLE>",out);

        printOut("<FORM METHOD=\"POST\" ACTION=\"/servlet/TDLServlet\">",out);
        printOut("<TT>Cache timeout (sec.): <I><INPUT TYPE=\"TEXT\" NAME=\"maxTime\" VALUE=\""+template.getTMaxTime()+"\" SIZE=8></I></TT><BR>",out);
        printOut("<TT>Cache maximum hits: <I><INPUT TYPE=\"TEXT\" NAME=\"maxHits\" VALUE=\""+template.getTMaxHits()+"\" SIZE=8></I></TT><BR>",out);
        printOut("<INPUT TYPE=\"SUBMIT\" VALUE=\"Apply\"><INPUT TYPE=HIDDEN NAME=\"todo\" VALUE=\"tcacheapp\"><INPUT TYPE=HIDDEN NAME=\"templ\" VALUE=\""+currentName+"\"><INPUT TYPE=HIDDEN NAME=\"extra\" VALUE=\""+extra+"\"></FORM><P>",out);

        printTail(out);
    }

    private void listTemplates (ServletOutputStream out) {
        Template t;
        printHead("",out);
        printOut("<CENTER><H2>Registered templates</H2></CENTER><HR>",out);
        printOut("<TABLE BORDER=1>",out);
        printOut("<TR><TD BGCOLOR=\"#FF8000\"><TT><B>Template name:</B></TT></TD>",out);
        printOut("<TD><TT><B>Template location:</B></TT></TD></TR>",out);
        Enumeration en = templateCollection.keys();
        while (en.hasMoreElements()) {
            String tname = (String)en.nextElement();
            t = (Template)templateCollection.get(tname);
            String loc = (String)t.getProperty(Constants.TemplateLocationProperty);
            printOut("<TR><TD BGCOLOR=\"#FF8000\"><TT>"+tname+"</TT></TD>",out);
            printOut("<TD><TT>"+loc+"</TT></TD></TR>",out);
        }
        printOut("</TABLE>",out);
        printTail(out);    
    }


    private void listProperties (ServletOutputStream out, int extra) {
        printHead("",out);
        printScript("24",out);
        printOut("<CENTER><H2>Template properties</H2></CENTER><HR>",out);
        printOut("<FORM NAME=\"forselect\"><SELECT NAME=\"tempname\" onChange=\"gotoPage()\">",out);
        Enumeration en = templateCollection.keys();
        int i=0;        
        String thisName = "";
        String currentName = "";
        while (en.hasMoreElements()) {
            thisName = (String)en.nextElement();
            if (i == extra) {
                printOut("<OPTION VALUE=\""+i+"\" SELECTED> "+thisName,out);
                currentName = thisName;
            }
            else {
                printOut("<OPTION VALUE=\""+i+"\"> "+thisName,out);
            }
            i++;
        }
        printOut("</SELECT></FORM><HR>",out);
        printOut("<TT><B>Properties for template : "+currentName+"</B></TT><P>",out);
        printTail(out);    
        Template template = (Template)templateCollection.get(currentName);
        printOut("<TABLE BORDER=1>",out);
        printOut("<TR><TD BGCOLOR=\"#FF8000\"><TT><B>Property:</B></TT></TD>",out);
        printOut("<TD><TT><B>Value:</B></TT></TD></TR>",out);

        en = template.getPropertyKeys();
        while (en.hasMoreElements()) {
            String pname = (String)en.nextElement();
            String pval = (String)template.getProperty(pname);
            if (!(pname.equals(Constants.IFLevelProperty)) && !(pname.equals(Constants.IFLevelConditionProperty)) ) {
                if (pname.equals(Constants.TemplateIDProperty) || pname.equals(Constants.ServletNameProperty) || pname.equals(Constants.ServletDirProperty) || pname.equals(Constants.TemplateLocationProperty)) {
                    printOut("<TR><TD BGCOLOR=\"#FF0000\"><TT>"+pname+"</TT></TD>",out);
                }
                else {
                    printOut("<TR><TD BGCOLOR=\"#FF8000\"><TT>"+pname+"</TT></TD>",out);
                }
                printOut("<TD><TT>"+pval+"</TT></TD></TR><P>",out);
            }
        }
        printOut("</TABLE><P>",out);
        if (!(template.isPropLoaded())) printOut("<TT>User defined properties not loaded.</TT>",out);
        printTail(out);    
                   
    }


	private void returnConfigPage (int pageCode, String serverName, ServletOutputStream out, int extra) {
	    switch (pageCode) {
	        case 11: servletInfo (serverName,out);break;
	        case 12: listKnownTags (out);break;
	        case 21: listTemplates (out);break;
	        case 22: break;
	        case 23: break;
	        case 24: listProperties(out,extra);break;
	        case 31: cacheInfo(out);break;
	        case 32: outputCacheControl(out);break;
	        case 33: templateCacheControl(out, extra);break;
	    }
	}
	
	
	private void showProperties(Template t, ServletOutputStream out) {
	    Enumeration enum = t.getPropertyKeys();
	    String key;
	    while (enum.hasMoreElements()) {
	        key = (String)enum.nextElement();
	        printToOutStream(key+" @ "+(String)t.getProperty(key)+"<BR>", out);
	    }
	}

	
}